const TABLE_NAME = process.env.TABLE_NAME!;
const REGION = process.env.REGION!;
const IOT_ENDPOINT = process.env.IOT_ENDPOINT!;

import { Handler } from "aws-lambda";
import {
  DynamoDBClient,
  GetItemCommand,
  GetItemCommandInput,
  PutItemCommand,
  PutItemCommandInput,
  ResourceNotFoundException,
} from "@aws-sdk/client-dynamodb";
import { marshall } from "@aws-sdk/util-dynamodb";
import axios from "axios";
const Parser = require("binary-parser").Parser;
import * as AWS from "aws-sdk";

const dbClient = new DynamoDBClient({ region: REGION }); // SDK v3

const iotdata = new AWS.IotData({
  endpoint: IOT_ENDPOINT,
  region: REGION,
});

export const addPayloadInternal = async (event: any) => {
  // parse WirelessDeviceId
  if (
    event.PayloadData == null ||
    event.PayloadData == "" ||
    event.WirelessMetadata.LoRaWAN.FPort !== 2
  ) {
    console.log(`no data.`);
    return;
  }
  const parsedData = parsePayload(event);
  const dataFromTable = await getItemFromTable(parsedData.wirelessDeviceId);

  const dataType = parsedData.dataType; // we can always get datatype
  if (dataType !== "gnss" && dataType !== "wifi") {
    console.log(
      `this is not gnss/wifi type. type is ${dataFromTable.dataType} ${parsedData.dataType}`,
    );
    return;
  }

  const newHex = dataFromTable.gnssData + parsedData.gnssData;

  const newAPArray = [
    ...dataFromTable.wifiData,
    ...(parsedData.wifiData ?? []),
  ];

  if (parsedData.counter === 0) {
    // all payload are filled.
    let latitude, longitude;
    if (dataType === "gnss") {
      const gnssHex = newHex.substring(0, dataFromTable.gnssDataLength * 2);
      [latitude, longitude] = await getLocation(
        gnssHex,
        parsedData.devEui,
        new Date().getTime() / 1000,
      );
      console.log(`gnss location: ${latitude} ${longitude}`);
    } else if (dataType === "wifi") {
      [latitude, longitude] = await getLocationByWifi(
        newAPArray,
        dataFromTable.numOfAp,
      );
      const trackerName = parsedData.devEui;
      console.log(`wifi location: ${latitude} ${longitude}`);
    }

    if (latitude && longitude) {
      await iotdata
        .publish({
          topic: `dt/${parsedData.devEui}/location`,
          payload: JSON.stringify({
            latitude: latitude,
            longitude: longitude,
            type: dataType,
          }),
          qos: 1,
        })
        .promise();
    }
    await resetDbState(parsedData.wirelessDeviceId);
    // TODO send it to Location Service.
  } else if (parsedData.counter > 0) {
    // put payload to table with ttl.
    if (
      dataFromTable.counter === -1 ||
      dataFromTable.counter - parsedData.counter === 1
    ) {
      await updateItem(
        parsedData.wirelessDeviceId,
        dataType === "gnss"
          ? parsedData.gnssDataLength
          : dataFromTable.gnssDataLength,
        parsedData.counter,
        newHex,
        dataType === "wifi" ? parsedData.numOfAp : dataFromTable.numOfAp,
        newAPArray,
      );
    } else {
      console.error(
        `invalid state counter : ${parsedData.counter}, last_counter : ${dataFromTable.counter} `,
      );
      await resetDbState(parsedData.wirelessDeviceId);
    }
    return;
  } else {
    console.error(`invalid counter: ${parsedData.counter}`);
  }
};

export const addPayload: Handler = async (event) => {
  //console.log(JSON.stringify(event));
  return addPayloadInternal(event);
};

interface ParsedData {
  wirelessDeviceId: string;
  counter: number;
  gnssDataLength: number;
  gnssData: string;
  numOfAp: number;
  wifiData: Array<number>;
  dataType: string;
  devEui: string;
}

async function resetDbState(deviceId: string) {
  await updateItem(deviceId, -1, -1, "", -1, []);
}

/**
 *
 * @param payload base64 encoded.
 * @returns deviceId, type, counter, size, dataHex, deveui.
 */
export function parsePayload(payload: any): ParsedData {
  const payloadData = payload.PayloadData;
  if (payloadData.length < 7) {
    throw new Error("payload is too small");
  }
  const buf = Buffer.from(payloadData, "base64");
  //console.log(`payload buffer: ${buf.toString("hex")}`);
  const parser = new Parser()
    .endianness("little")
    .int8("type")
    .choice("data", {
      tag: "type",
      choices: {
        83: Parser.start() // temp/humid
          .endianness("little")
          .uint16("power")
          .uint16("temperature")
          .uint16("humidity"),
        71: Parser.start() // gnss
          .endianness("little")
          .uint8("counter")
          .string("id", { length: 2 })
          .uint16("gnssDataLength")
          .uint32("radioTiming")
          .uint32("comutationTiming")
          .uint16("delaySinceCapture")
          .seek(-14)
          .string("gnssdata", {
            encoding: "hex",
            length: 47,
          }),
        87: Parser.start() // wifi
          .endianness("little")
          .uint8("counter")
          .string("id", { length: 2 })
          .uint8("dummy")
          .uint8("numOfAp")
          .uint32("timing1")
          .uint32("timing2")
          .uint32("timing3")
          .uint32("timing4")
          .seek(-20)
          .array("wifidata", {
            type: "uint8",
            length: 47,
          }),
      },
    });
  let result = parser.parse(buf);

  // remove header from data if it is initial packet.
  if (result.data.id === "@G") {
    result.data.gnssdata = result.data.gnssdata.substring(
      28,
      result.data.gnssdata.length,
    );
  } else if (result.data.id === "@W") {
    result.data.wifidata = result.data.wifidata.slice(
      20,
      result.data.wifidata.length,
    );
  }

  // dataType
  let dataType = "unknown";
  if (result.type == 71) {
    dataType = "gnss";
  } else if (result.type == 87) {
    dataType = "wifi";
  } else {
    dataType = result.type.toString();
  }

  return {
    wirelessDeviceId: payload.WirelessDeviceId,
    counter: result.data.counter,
    gnssDataLength: result.data.gnssDataLength,
    gnssData: result.data.gnssdata,
    numOfAp: result.data.numOfAp,
    wifiData: result.data.wifidata,
    dataType: dataType,
    devEui: payload.WirelessMetadata.LoRaWAN.DevEui,
  };
}

export async function getLocation(
  gnssHexData: string,
  devEui: string,
  captureTime: number,
): Promise<[number | null, number | null]> {
  let counter = 0;
  const devEuiWithHyphen = [...devEui]
    .map((char) => {
      return counter++ % 2 == 1 ? `${char}-` : char;
    })
    .join("")
    .replace(/-$/, "");
  const resp = await axios.post(
    "https://mgs.loracloud.com/api/v1/device/send",
    {
      deveui: devEuiWithHyphen, //like "70-b3-d5-7e-d0-05-00-33",
      uplink: {
        msgtype: "gnss",
        payload: gnssHexData,
        gnss_capture_time: captureTime,
      },
    },
    {
      headers: {
        Authorization: await getLoRaCloudTokenFromSM(
          "prototyping/lorawan/loracloud",
        ),
        "Content-Type": "application/json",
      },
    },
  );
  //console.log(resp.data.result);
  const position = resp.data.result.position_solution?.llh;
  if (position) {
    return [position[0], position[1]];
  } else {
    return [null, null];
  }
}

async function getLocationByWifi(
  newAPArray: number[],
  numOfAp: number,
): Promise<[number | null, number | null]> {
  let buf = Buffer.from(newAPArray);
  // console.log(buf);
  const parser = new Parser()
    .endianness("little")
    .string("macAddress", {
      encoding: "hex",
      length: 6,
    })
    .seek(1) //.uint8("channel")
    .seek(1) //.uint8("type")
    .uint8("signalStrength");
  const wifiInfo = [];
  for (let i = 0; i < numOfAp; i++) {
    let apInfo = parser.parse(buf);
    apInfo.signalStrength -= 256;
    let counter = 0;
    apInfo.macAddress = [...apInfo.macAddress]
      .map((char) => {
        return counter++ % 2 == 1 ? `${char}:` : char;
      })
      .join("")
      .replace(/:$/, "");

    wifiInfo.push(apInfo);
    buf = buf.slice(9, buf.length);
  }

  // console.log(`wifiInfo ${JSON.stringify(wifiInfo)}`);

  const resp = await axios.post(
    "https://mgs.loracloud.com/api/v1/solve/loraWifi",

    {
      lorawan: [
        {
          gatewayId: "00",
          rssi: 0,
          snr: 0,
          toa: 0,
          antennaId: 0,
          antennaLocation: { latitude: 0, longitude: 0, altitude: 0 },
        },
      ],
      wifiAccessPoints: wifiInfo,
    },
    {
      headers: {
        Authorization: await getLoRaCloudTokenFromSM(
          "prototyping/lorawan/loracloud",
        ),
        "Content-Type": "application/json",
      },
    },
  );
  const result = resp.data.result;
  return [result.latitude, result.longitude];
}

/**
 *
 * @param deviceId
 * @returns counter , payload
 */
export async function getItemFromTable(deviceId: string): Promise<ParsedData> {
  const dataFromTable: ParsedData = {
    wirelessDeviceId: deviceId,
    counter: -1,
    gnssDataLength: -1,
    gnssData: "",
    numOfAp: -1,
    wifiData: [],
    dataType: "",
    devEui: "",
  };
  try {
    const input: GetItemCommandInput = {
      TableName: TABLE_NAME,
      Key: marshall({
        WirelessDeviceId: deviceId,
      }),
    };
    const command = new GetItemCommand(input);
    const resp = await dbClient.send(command);

    // console.log(resp.Item);
    if (resp.Item == null) {
      return dataFromTable;
    }

    dataFromTable.gnssData = resp.Item["gnssData"]["S"] ?? "";
    dataFromTable.counter = parseInt(resp.Item["counter"]["N"] ?? "-1");
    dataFromTable.gnssDataLength = parseInt(
      resp.Item["gnssDataLength"]["N"] ?? "-1",
    );
    dataFromTable.numOfAp = parseInt(resp.Item["numOfAp"]["N"] ?? "-1");
    dataFromTable.wifiData = JSON.parse(resp.Item["wifiData"]["S"] ?? "[]"); // TODO

    return dataFromTable;
  } catch (error) {
    console.log(error);
  }
  return dataFromTable;
}

export async function updateItem(
  deviceId: string,
  gnssDataLength: number,
  counter: number,
  newHex: string,
  numOfAp: number,
  wifiData: Array<number>,
) {
  try {
    const input: PutItemCommandInput = {
      TableName: TABLE_NAME,
      Item: marshall({
        WirelessDeviceId: deviceId,
        gnssDataLength: gnssDataLength,
        counter: counter,
        gnssData: newHex,
        numOfAp: numOfAp,
        wifiData: JSON.stringify(wifiData),
        ttl: new Date().getTime() / 1000 + 60,
      }),
    };
    const command = new PutItemCommand(input);
    const resp = await dbClient.send(command);

    // console.log(resp);
  } catch (error) {
    console.log(error);
  }
}
export async function getLoRaCloudTokenFromSM(
  secretId: string,
): Promise<string> {
  const client = new AWS.SecretsManager({
    region: REGION,
  });
  const out = await new Promise<string>((resolve, reject) => {
    client.getSecretValue(
      {
        SecretId: secretId,
      },
      (err, data) => {
        if (err) {
          reject(err);
        } else if ("SecretString" in data) {
          const secretObj = JSON.parse(data.SecretString!);
          if ("LoRaCloudToken" in secretObj) {
            resolve(secretObj.LoRaCloudToken);
          } else reject(new Error("Token not found"));
        } else reject(new Error("SecretString not found"));
      },
    );
  });
  return out;
}
