import { Stack, StackProps } from "aws-cdk-lib";
import { Construct } from "constructs";
import { NodejsFunction } from "aws-cdk-lib/aws-lambda-nodejs";
import * as dynamodb from "aws-cdk-lib/aws-dynamodb";
import * as iot from "@aws-cdk/aws-iot-alpha";
import * as actions from "@aws-cdk/aws-iot-actions-alpha";
import * as secretsmanager from "aws-cdk-lib/aws-secretsmanager";
import { custom_resources } from "aws-cdk-lib";
import * as iam from "aws-cdk-lib/aws-iam";
export class LoracloudStack extends Stack {
  constructor(scope: Construct, id: string, props?: StackProps) {
    super(scope, id, props);
    const region_account = `${process.env.CDK_DEFAULT_REGION}:${process.env.CDK_DEFAULT_ACCOUNT}:`;

    // ddb
    const loraCloudPayloadCacheTable = new dynamodb.Table(
      this,
      "loraCloudPayloadCache",
      {
        partitionKey: {
          name: "WirelessDeviceId",
          type: dynamodb.AttributeType.STRING,
        },
        stream: dynamodb.StreamViewType.NEW_AND_OLD_IMAGES,
        encryption: dynamodb.TableEncryption.AWS_MANAGED,
        timeToLiveAttribute: "ttl",
        pointInTimeRecovery: true,
      },
    );

    // secretsmanager
    const secret = new secretsmanager.Secret(this, "loracloudtoken", {
      secretName: `prototyping/lorawan/loracloud`,
      generateSecretString: {
        secretStringTemplate: JSON.stringify({
          LoRaCloudToken: "",
        }),
        generateStringKey: "LoRaCloudToken",
      },
    });

    const addPayloadFunction = new NodejsFunction(
      this,
      `LoRaCloudPayloadHandler_addPayload`,
      {
        entry: "lambda/lora-cloud.ts",
        handler: "addPayload",
      },
    );
    secret.grantRead(addPayloadFunction);
    loraCloudPayloadCacheTable.grantReadWriteData(addPayloadFunction);
    addPayloadFunction.addEnvironment(
      "TABLE_NAME",
      loraCloudPayloadCacheTable.tableName,
    );

    const topic = "dt/lorawan";
    new iot.TopicRule(this, "LoRaWANTopicRule", {
      sql: iot.IotSql.fromStringAsVer20160323(
        `SELECT *, timestamp() as timestamp FROM '${topic}'`,
      ),
      actions: [new actions.LambdaFunctionAction(addPayloadFunction)],
    });

    const getIoTEndpoint = new custom_resources.AwsCustomResource(
      this,
      "IoTEndpoint",
      {
        onCreate: {
          service: "Iot",
          action: "describeEndpoint",
          physicalResourceId:
            custom_resources.PhysicalResourceId.fromResponse("endpointAddress"),
          parameters: {
            endpointType: "iot:Data-ATS",
          },
        },
        policy: custom_resources.AwsCustomResourcePolicy.fromSdkCalls({
          resources: custom_resources.AwsCustomResourcePolicy.ANY_RESOURCE,
        }),
      },
    );

    const IOT_ENDPOINT = getIoTEndpoint.getResponseField("endpointAddress");
    addPayloadFunction.addEnvironment("IOT_ENDPOINT", IOT_ENDPOINT);
    addPayloadFunction.addToRolePolicy(
      new iam.PolicyStatement({
        effect: iam.Effect.ALLOW,
        resources: [`arn:aws:iot:${region_account}topic/*`],
        actions: ["iot:Publish"],
      }),
    );
  }
}
